#include "svv.h"
#include "debug.h"
#include <windows.h>
#include <stdio.h>
#include <psapi.h>
#include <new>


//#define ALLOW_SINGLE_BYTE_KERNEL_MODIFICATIONS


void dumpBytes (DWORD startAddr, DWORD len) {

	printf ("dumping [%x-%x]:", startAddr, startAddr+len);
	for (DWORD a = startAddr; a <= startAddr + len; a++) 
		printf ("%02x ", (*((PBYTE)a)) & 0xff);
	printf ("\n");

}	

Module::Module () {
	importantModule = false;
	kernelModule = false;
	isNtoskrnl = false;
	base = 0;
	size = 0;
	name[0] = 0;
	path[0] = 0;

	hImage = NULL;

	moduleImageHidden = false;
	moduleImageReadError = false;
	moduleImageNotPresent = false;
	errCode = ERROR_SUCCESS;

	
	includeInReport = false;

	verdictLevel = INVALID_VERDICT;

	kernelAgent = NULL;


	noOfSections = 0;

	discrBlocksHead = NULL;
	exclusionFilter = new (std::nothrow) ExclusionFilter (this);

	diagnosticMode = false;
}

Module::Module (DWORD moduleBase) {
	Module::Module();
	base = moduleBase;
}

Module::Module (WCHAR *moduleName, WCHAR *modulePath, DWORD moduleBase, DWORD moduleSize) {
	Module::Module();
	base = moduleBase;
	size = moduleSize;
	wcsncpy (name, moduleName, sizeof (name)/sizeof(WCHAR));
	wcsncpy (path, modulePath, sizeof (path)/sizeof(WCHAR));
	
}

Module::~Module() {
	// free discrepancies list
	for (DWORD i = 0; i < noOfSections; i++) {

		delete sections[i].sectionInFile;
		delete sections[i].sectionInMemory;
	}



	struct DiscrepancyBlock *discrBlk = discrBlocksHead, *next;
	if (discrBlk) do {
		next = discrBlk->next;
		delete discrBlk;
		discrBlk = next;
	} while (discrBlk);

	delete exclusionFilter;

	closeModuleImageFile ();
	
}




Module::ExclusionFilter::ExclusionFilter (Module *m) {
	module = m;
	filterUsedForRtlPrefetchMemoryNonTemporal = false;
	filterUsedForKeFlushCurrentTb_SingleByte = false;
	filterUsedForKiSystemCallExitBranch = false;

	lastMatchDescription = NULL;
}

bool Module::ExclusionFilter::matchFilter (struct Module::Section *section, DWORD offsetIntoCodeSection, DWORD len) {
	DWORD offset = offsetIntoCodeSection;
	lastMatchDescription = L"???";

	//if ((len == 1) && (section->sectionInMemory[offset] == 0xcc)) {
	//	lastMatchDescription = L"INT3 (breakpoint)";
	//	return true;
	//}

	if (!wcscmp (module->name, L"hal.dll")) {
		// we generally don't care much about HAL, but anyway :)

		for (DWORD i = 0; i < len; i++) 
			if (section->sectionInMemory[offset + i] != 0) return false;
		lastMatchDescription = L"typical HAL customization";
		return true;
	}	

//	if (!wcscmp (module->name, L"ntoskrnl.exe")) {
	if (module->isNtoskrnl) {
		
		// allow KeFlushCurrentTb() changes (All Windows 2000, XP, 2003)
		BYTE KeFlushCurrentTbOrigBytes[] =
			"\xd8\x0f\x22\xd8\xc3\x0f\x20\xe0\x25\x7f\xff\xff\xff\x0f\x22\xe0\x0d\x80";

		BYTE KeFlushCurrentTbNewBytes[] = 
			"\xe0\x25\x7f\xff\xff\xff\x0f\x22\xe0\x0d\x80\x00\x00\x00\x0f\x22\xe0\xc3";

		if (!memcmp (
				&section->sectionInFile[offset],
				KeFlushCurrentTbOrigBytes,
				sizeof (KeFlushCurrentTbOrigBytes)) &&
			!memcmp (
				&section->sectionInMemory[offset],
				KeFlushCurrentTbNewBytes, 
				sizeof (KeFlushCurrentTbNewBytes))
			) {
				lastMatchDescription = L"KeFlushCurrentTb()";
				return true;
			}

#ifdef ALLOW_SINGLE_BYTE_KERNEL_MODIFICATIONS
		lastMatchDescription = L"single byte modification";
		if (len == 1) return true;
		
#else

		// KeFlushCurrentTb - single byte (All Windows 2000, XP, 2003)
		if (section->sectionInFile[offset] == 0xc3 &&
			section->sectionInMemory[offset] == 0x00 &&
			!filterUsedForKeFlushCurrentTb_SingleByte) {
				filterUsedForKeFlushCurrentTb_SingleByte = true;
				lastMatchDescription = L"KeFlushCurrentTb() [c3->00]";
				return true;
			}

		// RtlPrefetchMemoryNonTemporal (All Windows 2000, XP, 2003)
		if (section->sectionInFile[offset] == 0xc3 &&
			section->sectionInMemory[offset] == 0x90 &&
			!filterUsedForRtlPrefetchMemoryNonTemporal) {
				filterUsedForRtlPrefetchMemoryNonTemporal = true;
				lastMatchDescription = L"RtlPrefetchMemoryNonTemporal() [c3->90]";
				return true;
			}

		// KiSystemCallExitBranch (Windows XP SP0 - SP2)
		if (section->sectionInFile[offset] == 0x05 &&
			section->sectionInMemory[offset] == 0x06 &&
			!filterUsedForKiSystemCallExitBranch) {
				filterUsedForKiSystemCallExitBranch = true;
				lastMatchDescription = L"KiSystemCallExitBranch() [05->06]";
				return true;
			}
#endif		
		return false;
	}
	return false;
}

bool Module::isJMPingCode (PBYTE code, DWORD len, DWORD eip, DWORD *targetAddr) {

	if (len >= 1 /* 5 */ && (code[0] == 0xe9 || code[0] == 0xe8)) {
		// JMP rel32 | CALL rel32
		DWORD rel32 = *((DWORD*)&code[1]);
		*targetAddr = eip + rel32;
		return true;
	}


	if (len >= 2 /* 6 */ && code[0] == 0xff && (code[1] == 0x25 || code[1] == 0x15)) {
		// JMP m32 | CALL m32

		DWORD m32 = *((DWORD*)&code[2]);
		*targetAddr = m32;
		return true;
	}


	return false;
}



bool Module::verifySection (struct Section* s) {

	_debug ("verifing section: %s, RVA = %#x, offs = %#x, sz = %d", s->name, s->RVA, s->fileOffs, s->size);
	DWORD sectionSz = s->hdr->Misc.VirtualSize;

	s->sectionInFile = new (std::nothrow) BYTE [sectionSz];
	if (!s->sectionInFile ) {
		_debug ("no memory for buffer for code s->hdr body!");
		SetLastError (ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}

	SetFilePointer (hImage, s->hdr->PointerToRawData, 0, FILE_BEGIN);
	DWORD bRead;
	if (!ReadFile (hImage, s->sectionInFile, sectionSz, &bRead, NULL)) {
		return false;
	}


	if (pefile.ntHdr.OptionalHeader.ImageBase != base) {
		bool ret = pefile.relocBuffer (
			(DWORD)s->sectionInFile,
			(DWORD)s->sectionInFile + sectionSz - 1,
			s->hdr->VirtualAddress, 
			base - pefile.ntHdr.OptionalHeader.ImageBase);
		if (!ret) {
			_debug ("relocation failed!");
			SetLastError (ERROR_BAD_FORMAT);
			return false;
		}
	}


	s->sectionInMemory = new (std::nothrow) BYTE [sectionSz];

	if (!s->sectionInMemory ) {
		_debug ("no memory for buffer for code s->hdr body!");
		SetLastError (ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}

	
	bool ret = kernelAgent->readVirtualMemory (
		s->hdr->VirtualAddress + base,
		s->sectionInMemory,
		sectionSz,
		&bRead);
	
	if (!ret || bRead != sectionSz) {
		SetLastError (ERROR_INVALID_ADDRESS);	
		return false;
	}

	// now compare the two buffers
	bool insideDiffBlock = false;
	DWORD diffsFound = 0;
	DWORD diffsNotAllowedByExclusionFiltrer = 0;
	DWORD diffRangeStart = 0;

	DWORD iatStart = pefile.IATStartRva + base;
	DWORD iatEnd = pefile.IATEndRva + base;

	DWORD eatStart = pefile.EATStartRva + base;
	DWORD eatEnd = pefile.EATEndRva + base;
	

	for (DWORD offset = 0; offset < sectionSz; offset++) {
		DWORD va = offset + base + s->hdr->VirtualAddress;

		// skip .idata which is typically merged with .text
		if (va >= iatStart && va <= iatEnd) continue;	

		if (!insideDiffBlock && 
			(s->sectionInFile[offset] != s->sectionInMemory[offset])) {		
			
			diffsFound++;
			insideDiffBlock = true;		

			diffRangeStart = offset;

			if (kernelAgent->isAddrInsideServiceTbl (base + s->hdr->VirtualAddress + offset)) {
				// modifications in SDT always report in 4-byte chunks
				
				diffRangeStart = (offset/sizeof(DWORD)) * sizeof (DWORD);
				struct DiscrepancyBlock* discrBlk = new (std::nothrow) struct DiscrepancyBlock;
				if (discrBlk == NULL) {
					SetLastError (ERROR_NOT_ENOUGH_MEMORY);
					return false;
				}
				memset (discrBlk, 0, sizeof (struct DiscrepancyBlock));
				discrBlk->type = SDT;
				discrBlk->section = s;
				discrBlk->address = diffRangeStart;
				discrBlk->len = sizeof (DWORD);
				discrBlk->fixable = true;

				discrBlk->modification.SDT.ServTblEntrNo = 
					(diffRangeStart + base + s->hdr->VirtualAddress 
					- kernelAgent->origKiServiceTblStart)/sizeof (DWORD);

				discrBlk->targetAddr = 
					*((DWORD*)(&s->sectionInMemory[diffRangeStart]));

				offset = diffRangeStart + sizeof (DWORD) - 1;

				diffsNotAllowedByExclusionFiltrer++;
				// insert at the end (sorted by address)
				struct DiscrepancyBlock* p = discrBlocksHead;
				if (!p) discrBlocksHead = discrBlk;
				else {
					for (; p->next; p = p->next);			
					p->next = discrBlk;
				}
				insideDiffBlock = false;
				continue;
			}
		}



		if (insideDiffBlock && (s->sectionInFile[offset] == s->sectionInMemory[offset])) {

			if ( (s->sectionInFile[offset + 1] != s->sectionInMemory[offset + 1]) ||
				 (s->sectionInFile[offset + 2] != s->sectionInMemory[offset + 2])) continue;
					// avoid fragmentation!

			DWORD diffsNo = offset - diffRangeStart;

			struct DiscrepancyBlock* discrBlk = new (std::nothrow) struct DiscrepancyBlock;
			if (discrBlk == NULL) {
				SetLastError (ERROR_NOT_ENOUGH_MEMORY);
				return false;
			}
			memset (discrBlk, 0, sizeof (struct DiscrepancyBlock));
			discrBlk->section = s;
			discrBlk->type = CODESECTION;
			discrBlk->address = diffRangeStart;
			discrBlk->len = diffsNo;
			discrBlk->fixable = true;
									
			if (exclusionFilter->matchFilter (s, discrBlk->address, discrBlk->len)) {
				discrBlk->modification.code.exlucsionDescription = exclusionFilter->lastMatchDescription;
				discrBlk->modification.code.excludedByFilter = true;
			}
			else  {
				diffsNotAllowedByExclusionFiltrer++;
				DWORD targetAddr = 0;
				DWORD startVA = base + s->hdr->VirtualAddress + diffRangeStart;

				if (isJMPingCode (&s->sectionInMemory[diffRangeStart], discrBlk->len, 
					startVA, &targetAddr)) {
					discrBlk->modification.code.isJMPingCode = true;

					discrBlk->targetAddr = targetAddr;
				}

	
			if (
					kernelAgent->isAddrInsideServiceTbl (startVA) &&
					kernelAgent->isAddrInsideServiceTbl (startVA + discrBlk->len))
					discrBlk->type = SDT;
				
			}

			if (va >= eatStart && va < eatEnd) discrBlk->type = EAT;
			
			
			if (pefile.findNearestExportedSymbol (
				s->hdr->VirtualAddress + diffRangeStart,
				64, &discrBlk->offsetFromSymbol,
				discrBlk->exportedName, sizeof (discrBlk->exportedName)
				)) {
					discrBlk->nearExportedSymbol = true;
			}


			// insert at the end (sorted by address)
			struct DiscrepancyBlock* p = discrBlocksHead;
			if (!p) discrBlocksHead = discrBlk;
			else {
				for (; p->next; p = p->next);			
				p->next = discrBlk;
			}
			insideDiffBlock = false;
			continue;
		}
	}
	_debug ("module %S, section %s: diffsFound = %d, diffsNotAllowedByExclusionFiltrer = %d",
		getName(), s->name, diffsFound, diffsNotAllowedByExclusionFiltrer);
	
	return true;
}


void Module::verify () {
	if (!kernelAgent) {
		_debug ("kernelAgent is NULL!");
		errCode = ERROR_BAD_ARGUMENTS;	
		return;
	}

	
	if (wcslen (path) == 0) {
		errCode = ERROR_BAD_ARGUMENTS;
		return;
	}

	if (!openModuleImageFile()) {
		moduleImageNotPresent = true;
		errCode = GetLastError();
		return;
	}


	pefile.setFileHandle (hImage);
	if (!pefile.load ()) {
		moduleImageReadError = true;
		errCode = GetLastError();
		return;
	}
	
	if (pefile.ntHdr.OptionalHeader.ImageBase != base) {
		bool ret = pefile.loadRelocTable();
		if (!ret) {
			_debug ("cannot load .reloc section");
			moduleImageReadError = true;
			errCode = ERROR_BAD_FORMAT;
			return;
		}
	}


	DWORD sctNo = 0;
	for (DWORD i = 0; i < pefile.noOfSections; i++) {

		DWORD chr = pefile.sectHdrs[i].Characteristics;
		if (chr & IMAGE_SCN_MEM_EXECUTE && 
			chr & IMAGE_SCN_CNT_CODE &&
			!(chr & IMAGE_SCN_LNK_REMOVE) &&
			!(chr & IMAGE_SCN_MEM_NOT_CACHED) &&
			!(chr & IMAGE_SCN_MEM_DISCARDABLE)) {
			
			sections[sctNo].hdr = &pefile.sectHdrs[i]; 
			sections[sctNo].RVA = pefile.sectHdrs[i].VirtualAddress;
			sections[sctNo].fileOffs = pefile.sectHdrs[i].PointerToRawData;
			sections[sctNo].size = pefile.sectHdrs[i].Misc.VirtualSize;

			sections[sctNo].sectionInFile = NULL;
			sections[sctNo].sectionInMemory = NULL;
			strncpy (sections[sctNo].name, (char*)pefile.sectHdrs[i].Name, 8);
			sections[sctNo].name[8] = 0;

			//if (isNtoskrnl && !strcmp (sections[sctNo].name, "PAGEVRFY")) {
			//	_debug("skiping PAGEVRFY section....\n");
			//	continue;
			//}



			if (isNtoskrnl) {
				bool skip = true;	
				if (!strcmp (sections[sctNo].name, ".text")) skip = false;
				if (!strcmp (sections[sctNo].name, "PAGE")) skip = false;
				if (skip) {
					_debug("skiping %s section....\n", sections[sctNo].name);
					continue;
				}
				
			}

			if (diagnosticMode) {
				printf ("-> about to verify %s section...\n", sections[sctNo].name);
				Sleep (3000);
			}


			bool verifyOk = verifySection (&sections[sctNo]);
			if (!verifyOk) {
				errCode = GetLastError();
				noOfSections = sctNo;
				return;
			}
		
			sctNo++;
		
			}
	}

	noOfSections = sctNo;


	return;
}

bool Module::fix (SystemVerdict targetVerdict) {


	DWORD bytesFixed = 0;
	struct DiscrepancyBlock *discrBlk = discrBlocksHead;
	for (;discrBlk; discrBlk = discrBlk->next) {
			
		if (discrBlk->verdictLevel <= targetVerdict) continue;
		if (!discrBlk->fixable) continue;
		DWORD addr = discrBlk->address + base + discrBlk->section->RVA;

		bool ret;
		ret = kernelAgent->writeVirtualMemory (
			addr,				
			&discrBlk->section->sectionInFile[discrBlk->address],
			discrBlk->len);
		if (!ret) return false;

		PBYTE buf = new (std::nothrow) BYTE [discrBlk->len];
		if (!buf) {
			SetLastError (ERROR_NOT_ENOUGH_MEMORY);
			return false;
		}

		DWORD bRead;
		ret = kernelAgent->readVirtualMemory (
			addr,
			buf,
			discrBlk->len,
			&bRead);
		
		if (!ret || bRead != discrBlk->len) {
			SetLastError (ERROR_INVALID_ADDRESS);	
			delete buf;
			return false;
		}


		if (memcmp (buf, &discrBlk->section->sectionInFile[discrBlk->address], discrBlk->len)!=0) {
			_debug ("repair failed: written bytes != read bytes!");
			SetLastError (ERROR_WRITE_FAULT);
		
			delete buf;
			return false;
		}
		delete buf;

		bytesFixed += discrBlk->len;
	
	}

	return true;
}




bool Module::fixInAllProcesses (SystemVerdict targetVerdict) {
	DWORD nElements = 0;
	DWORD *pPidsArray = NULL;
	DWORD cbReturned;
	bool problems = false;
	bool enumProblems = false;
		
	do {
		nElements += 100; 
		delete pPidsArray;
		pPidsArray = new (std::nothrow) DWORD [nElements];
		if (pPidsArray == NULL) {
			_debug ("Cannot allocate memory for userland processes list!");
			SetLastError (ERROR_NOT_ENOUGH_MEMORY);		
			return false;
		}
		EnumProcesses (pPidsArray, nElements*sizeof(DWORD), &cbReturned);
	    
	} while (cbReturned/sizeof(DWORD) == nElements);
	
	int nProcesses = cbReturned/sizeof(DWORD);

	for (int i = 0; i < nProcesses; i++) {
		if (pPidsArray[i] == 0) continue; // skip process 0
		if (i == 1) continue; // skip System process, always second on the list
		if (pPidsArray[i] == GetCurrentProcessId()) continue;
		
		__try {
			if (!fixInForeignProcess (pPidsArray[i], targetVerdict)) {
				if (!enumProblems)
					_debug ("error code = %x (pid = %d), skipping...", 
					GetLastError(), pPidsArray[i]);
				problems = true;
			
			}
	
		} __except (EXCEPTION_EXECUTE_HANDLER) {
			_debug ("exception inside fixInForeignProcess(%d)!", pPidsArray[i]);
			problems = true;
		}
	}

	return !problems;
}

bool Module::fixInForeignProcess (DWORD pid, SystemVerdict targetVerdict) {
	HANDLE hProcess = OpenProcess (
		PROCESS_ALL_ACCESS,
		false,
		pid);
	if (hProcess == NULL) return false;
	
	// check if this module is present in this process:
	DWORD count = 100, neededSize;
	HMODULE *hMods = new (std::nothrow) HMODULE [count];
	BOOL ret;

	while (1) {
		ret = EnumProcessModules (
			hProcess,
			hMods,
			count * sizeof (HMODULE),
			&neededSize);

		if (!ret) {
			_debug ("EnumProcessModules failed for pid %d, code = %#x", pid, GetLastError());
			goto err;	
		}

		if (neededSize > (count * sizeof (HMODULE))) {
			delete hMods;
			hMods = new (std::nothrow) HMODULE [count *= 2];
			continue;
		} else break;
	};

	
	DWORD noOfModules = neededSize/sizeof (HMODULE);

	bool moduleFound = false;
	DWORD foreignBase = 0;
	
	for (DWORD i = 0;i < noOfModules; i++) {
		WCHAR modPath [MAX_PATH_LEN];
		ret = GetModuleFileNameEx (
			hProcess,
			hMods[i],
			modPath,
			sizeof (modPath));
	
		if (ret == 0) {
				DWORD err = GetLastError();
				_debug ("GetModuleFileName (base = %#x) failed with err = %#x", base, err);
				SetLastError (err);	
				goto err;
		}
		
		if (!wcscmp(modPath, path)) {
			moduleFound = true;
			foreignBase = (DWORD)hMods[i];
			break;
		}
	}
	delete []hMods;
	if (!moduleFound) {
		CloseHandle(hProcess);
		return true;	// this module is not present in the foreign process
	}

	DWORD bytesFixed = 0;
	struct DiscrepancyBlock *discrBlk = discrBlocksHead;
	for (;discrBlk; discrBlk = discrBlk->next) {
		if (discrBlk->verdictLevel <= targetVerdict) continue;	
		BOOL ret;
		DWORD oldProt, dummy;
		ret = VirtualProtectEx (
			hProcess,
			(LPVOID)(discrBlk->address + foreignBase + discrBlk->section->RVA),
			discrBlk->len,
			PAGE_READWRITE,
			&oldProt);
		if (!ret) {
			_debug ("VirtualProtectEx failed for pid %d, code = %#x", pid, GetLastError());
			goto err;
		}

		ret = WriteProcessMemory (
			hProcess, 
			(LPVOID)(discrBlk->address + foreignBase + discrBlk->section->RVA),
			&discrBlk->section->sectionInFile[discrBlk->address],
			discrBlk->len,
			NULL);

		if (!ret) {
			_debug ("WriteProcessMemory  failed for pid %d, code = %#x", pid, GetLastError());
			goto err;
		}

		ret = VirtualProtectEx (
			hProcess,
			(LPVOID)(discrBlk->address + foreignBase + discrBlk->section->RVA),
			discrBlk->len,
			oldProt,
			&dummy);
		if (!ret) goto err;


		bytesFixed += discrBlk->len;
		
	}

	
	return true;	

err:
	CloseHandle (hProcess);
	return false;
}

bool Module::isPathHidden (WCHAR *path) {

	WIN32_FIND_DATA FindFileData;

	WCHAR *slash = NULL;
	int maxsteps = 128;


	
	WCHAR *ptr = path;
	if (ptr[0] == '\\') ptr ++;
	while (slash = wcschr (ptr, '\\')) {
		*slash = 0;
		if (ptr[wcslen (ptr) - 1] == ':') {
			*slash = '\\';        
			ptr = slash + 1;
			continue; // skip c:
		}


		HANDLE hFind = FindFirstFile(path, &FindFileData);

		if (hFind == INVALID_HANDLE_VALUE) {
			_debug ("hidden dir: %S\n", path);
			
			*slash = '\\';        
			FindClose (hFind);
			return true;
		}

		FindClose (hFind);

		*slash = '\\';        
		ptr = slash + 1;
		if (!maxsteps--) break;

	}
	return false;
}

bool Module::openModuleImageFile() {

	if (hImage != NULL && hImage != INVALID_HANDLE_VALUE) 
		return true;	// already opened

	//_debug ("orig module path: %S", path);
	WCHAR newPath[MAX_PATH_LEN];
	wcscpy (newPath, path);
	if (wcschr (path,'\\') == NULL) 
		swprintf (newPath, L"%%SystemRoot%%\\System32\\Drivers\\%s", path);

	
	WCHAR *i;
	if ((i = wcsstr (path, L"\\SystemRoot\\"))) {
		WCHAR *j = wcschr (i + 1, '\\');
		if (j == 0) {
			SetLastError (ERROR_UNIDENTIFIED_ERROR);
			return false;
		}
		swprintf (newPath, L"%%SystemRoot%%\\%s", j + 1);
	}


	if ((i = wcsstr (path, L"\\??\\"))) {
		WCHAR *j = wcschr (i + 1, '\\');
		if (j == 0) {
			SetLastError (ERROR_UNIDENTIFIED_ERROR);
			return false;
		}
		swprintf (newPath, L"%s", j + 1);
	}

	WCHAR expandedPath [MAX_PATH];

expand:	
	DWORD ret = ExpandEnvironmentStrings(newPath,
		expandedPath, sizeof (expandedPath)/sizeof(WCHAR));
	if (ret == 0) return false;

	
	if (wcschr (expandedPath,':') == NULL) {
		swprintf (newPath, L"%%SystemDrive%%%s", expandedPath);
		goto expand;
	}
	

	moduleImageHidden = isPathHidden(expandedPath);

	//_debug ("trying open: %S", expandedPath);
	hImage = CreateFile (expandedPath,
		GENERIC_READ,
		FILE_SHARE_READ,
		NULL,
		OPEN_EXISTING,
		0,
		NULL);

	if (hImage == INVALID_HANDLE_VALUE) {
		moduleImageNotPresent = true;
		return false;
	}

	wcscpy (effectivePath, expandedPath);

	moduleImageNotPresent = false;
	
	return true;
}



