#include "pefile.h"
#include "debug.h"
#include <stdio.h>
#include <new>

PEFile::PEFile (HANDLE fileHandle) {
	hFile = fileHandle;
	sectHdrs = NULL;
	relocTable = NULL;
	relocTableSz = 0;

	IATStartRva = IATEndRva = 0;
	EATStartRva = EATEndRva = 0;

	EAT = NULL;
	ENT = NULL;
	EOT = NULL;

}

PEFile::~PEFile () {
	delete []sectHdrs;
	delete []relocTable;

	delete EAT;
	delete ENT;
	delete EOT;
}


bool PEFile::load() {


	DWORD bRead;
	SetFilePointer (hFile, 0, 0, FILE_BEGIN);
	BOOL ret = ReadFile (hFile, &dosHdr, sizeof (dosHdr), &bRead, NULL);
	if (!ret) return false;
	if (dosHdr.e_magic != 'ZM') {
		SetLastError (ERROR_BAD_EXE_FORMAT);
		return false;
	}

	SetFilePointer (hFile, dosHdr.e_lfanew, 0, FILE_BEGIN);
	
	ret = ReadFile (hFile, &ntHdr, sizeof (ntHdr), &bRead, NULL);
	if (!ret) return false;

	if (ntHdr.Signature != 0x00004550) {
		SetLastError (ERROR_BAD_EXE_FORMAT);
		return true;
	}

	noOfSections = ntHdr.FileHeader.NumberOfSections;
	sectHdrs = new (std::nothrow) IMAGE_SECTION_HEADER [noOfSections];

	for (DWORD i = 0; i < noOfSections; i++) {
		DWORD sectOffset = dosHdr.e_lfanew + 
			sizeof (IMAGE_NT_HEADERS32) +
			i * sizeof (IMAGE_SECTION_HEADER);
		
		SetFilePointer (hFile, sectOffset, 0, FILE_BEGIN);
		ret = ReadFile (hFile, &sectHdrs[i], sizeof (IMAGE_SECTION_HEADER), &bRead, NULL);
		if (!ret) return false;
	}

	_debug ("");
	
	_debug ("base: %#x", ntHdr.OptionalHeader.ImageBase);


	IATStartRva = 
		ntHdr.OptionalHeader.
		DataDirectory[IMAGE_DIRECTORY_ENTRY_IAT].VirtualAddress;
	
	IATEndRva = IATStartRva + ntHdr.OptionalHeader.
		DataDirectory[IMAGE_DIRECTORY_ENTRY_IAT].Size;
	
	DWORD base = ntHdr.OptionalHeader.ImageBase;

	_debug ("IAT RVA: %#x-%#x", IATStartRva, IATEndRva);


	IMAGE_SECTION_HEADER *sctContaingExports = 
		findSectionByRVA (ntHdr.OptionalHeader.
		DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].VirtualAddress);

	
	if (!sctContaingExports) {
		_debug ("No IMAGE_EXPORT_DIRECTORY");	
		memset (&exports, 0, sizeof (exports));
		return true; // this is not error
	}


	_debug ("DATA_DIR[EXPORTS] RVA = %#x, inside section: %s", ntHdr.OptionalHeader.
		DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].VirtualAddress, 
		sctContaingExports->Name);


	DWORD exportsOffs = ntHdr.OptionalHeader.
		DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].VirtualAddress -
		(sctContaingExports->VirtualAddress - sctContaingExports->PointerToRawData);
	
	

	SetFilePointer (hFile, exportsOffs, 0, FILE_BEGIN);
	ret = ReadFile (hFile, &exports, sizeof (IMAGE_EXPORT_DIRECTORY), &bRead, NULL);
	if (!ret) return false;


	EATStartRva = exports.AddressOfFunctions;
	EATEndRva = EATStartRva + exports.NumberOfFunctions * sizeof (DWORD);

	_debug ("EAT RVA (VA): %#x-%#x", EATStartRva, EATEndRva);

	// read in EAT table
	EAT = new (std::nothrow) DWORD [exports.NumberOfFunctions];
	if (!EAT) {
		SetLastError (ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}

	DWORD EATOffs =  EATStartRva -
		(sctContaingExports->VirtualAddress - sctContaingExports->PointerToRawData);

	SetFilePointer (hFile, EATOffs, 0, FILE_BEGIN);
	ret = ReadFile (hFile, EAT, exports.NumberOfFunctions * sizeof (DWORD), &bRead, NULL);
	if (!ret) return false;

	// read in ENT table
	ENT = new (std::nothrow) DWORD [exports.NumberOfNames];
	if (!EAT) {
		SetLastError (ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}

	DWORD ENTOffs =  exports.AddressOfNames -
		(sctContaingExports->VirtualAddress - sctContaingExports->PointerToRawData);

	SetFilePointer (hFile, ENTOffs, 0, FILE_BEGIN);
	ret = ReadFile (hFile, ENT, exports.NumberOfNames * sizeof (DWORD), &bRead, NULL);
	if (!ret) return false;

	// read in EOT table
	EOT = new (std::nothrow) WORD [exports.NumberOfNames];
	if (!EAT) {
		SetLastError (ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}

	DWORD EOTOffs =  exports.AddressOfNameOrdinals -
		(sctContaingExports->VirtualAddress - sctContaingExports->PointerToRawData);

	SetFilePointer (hFile, EOTOffs, 0, FILE_BEGIN);
	ret = ReadFile (hFile, EOT, exports.NumberOfNames * sizeof (WORD), &bRead, NULL);
	if (!ret) return false;

	return true;

}

bool PEFile::loadRelocTable () {
	IMAGE_SECTION_HEADER *relocSct = findSectionByName (".reloc");
	if (relocSct == NULL) {
		_debug ("No .reloc section found in image!");
		return false;
	}

	DWORD relocFileOffs = relocSct->PointerToRawData;
	relocTableSz = relocSct->Misc.VirtualSize;
	relocTable = new (std::nothrow) BYTE [relocTableSz];
	if (!relocTable) {
		_debug ("no memory for buffer for relocTable!");
		SetLastError (ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}

	SetFilePointer (hFile, relocFileOffs, 0, FILE_BEGIN);
	DWORD bRead;

	BOOL ret = ReadFile (hFile, relocTable, relocTableSz, &bRead, NULL);
	if (!ret) {
		_debug ("image read error while reading .reloc body (err = %#x)\n",
			GetLastError());
		return false;
	}
	return true;
}


IMAGE_SECTION_HEADER* PEFile::findSectionByName (char *sectName) {
	//size_t nameLen = (strlen(sectName) <= 8) ? strlen(sectName) : 8;
	for (DWORD i = 0; i < noOfSections; i++)
		if (!strncmp (sectName, (char*)sectHdrs[i].Name, 8))
			return &sectHdrs[i];
	return NULL;
}

IMAGE_SECTION_HEADER* PEFile::findSectionByRVA (DWORD RVA) {
	
	for (DWORD i = 0; i < noOfSections; i++)
		if (RVA >= sectHdrs[i].VirtualAddress &&
			RVA < sectHdrs[i].VirtualAddress + sectHdrs[i].Misc.VirtualSize)
			return &sectHdrs[i];

	return NULL;
}


bool PEFile::relocBuffer(DWORD start, DWORD end, DWORD sectionRVA, DWORD relOffset) {
  __try {
      for (IMAGE_BASE_RELOCATION *relEntry = (IMAGE_BASE_RELOCATION *)relocTable;
		(((DWORD)relEntry) < ((DWORD)relocTable + relocTableSz) &&
		relEntry->SizeOfBlock);
		relEntry = (IMAGE_BASE_RELOCATION *)(((DWORD)relEntry) + relEntry->SizeOfBlock)) {
		
			WORD *relWordTbl = (WORD*)(((DWORD)relEntry) + 0x8);
			int relWordTblSz = (relEntry->SizeOfBlock - 0x8)/sizeof (WORD);

			for (int i = 0; i < relWordTblSz; i++)  {
		
				DWORD addr = relEntry->VirtualAddress + 
					(relWordTbl[i] & 0x0fff) 
					+ start - sectionRVA;
				
      			if (addr < start || addr > end) continue;
	    
				switch ((relWordTbl[i] & 0xf000)>>12) {
					case IMAGE_REL_BASED_ABSOLUTE:
						break;
					case IMAGE_REL_BASED_HIGHLOW:
						*(DWORD*)(addr) += relOffset;
						break;
					default: break;      
				}   			
			} 	 
		
		}
	
  }	__except (EXCEPTION_EXECUTE_HANDLER) {
	  _debug ("Exception in relocBuffer ()!");
	  SetLastError (ERROR_BAD_FORMAT);
	  return false;
  }
  return true;
}

bool PEFile::RVA2fileOffset (DWORD RVA, DWORD *fileOffset) {
	IMAGE_SECTION_HEADER *sctContaingRVA = 
		findSectionByRVA (RVA);

	if (!sctContaingRVA) return false;

	*fileOffset = RVA -
		(sctContaingRVA->VirtualAddress - sctContaingRVA->PointerToRawData);
	return true;
}


bool PEFile::readBytes (DWORD offset, PBYTE buf, DWORD len) {
	SetFilePointer (hFile, offset, 0, FILE_BEGIN);
	DWORD bRead = 0;
	BOOL ret = ReadFile (hFile, buf, len, &bRead, NULL);
	if (bRead != len) {
		_debug ("bRead != len (%d != %d)", bRead, len);
		return false;
	}
	if (ret) return true; else return false;
}

bool PEFile::findNearestExportedSymbol (DWORD RVA, DWORD maxDistance, DWORD *distance,
		char *exportedName, DWORD exportedNameMaxLen) {
			
			
	if (exports.AddressOfFunctions == 0) return false;	// no exports

	if (!EAT) return false;

	DWORD bestMatchDistance = maxDistance + 1;
	DWORD bestMatchNo = -1;
	for (DWORD i = 0; i < exports.NumberOfFunctions; i++)
		if (EAT[i] <= RVA && ((RVA - EAT[i]) < bestMatchDistance)) {
			bestMatchDistance = RVA - EAT[i];
			bestMatchNo = i;
		}

	if (bestMatchNo == -1) return false;
	
	DWORD ordinal = bestMatchNo + exports.Base;

	// now find the ordinal in the EOT

	DWORD ENTindex = -1;
	for (DWORD i = 0; i < exports.NumberOfNames; i++) 
		if (EOT[i] == (ordinal - exports.Base)) {
			ENTindex = i;
			break;
		}

	if (ENTindex == -1) {
		// exported by ordinal only
		char buf [16];
		sprintf (buf, "ordinal %d", ordinal, ENTindex);
		strncpy (exportedName, buf, exportedNameMaxLen);
		if (distance) *distance = bestMatchDistance;
		return true;
	}

	DWORD nameOffset = 0;
	bool offsetFound = RVA2fileOffset (ENT[ENTindex], &nameOffset);

	if (!offsetFound) return false;

	char buf[128];
	if (!(readBytes (nameOffset, (PBYTE)buf, sizeof (buf)))) return false;

	strncpy (exportedName, buf, exportedNameMaxLen);
	if (distance) *distance = bestMatchDistance;
	return true;
										
}

// return RVA of the exported symbol
DWORD PEFile::findExportedSymbol (const char *symName) {
	if (exports.AddressOfFunctions == 0) return NULL;	// no exports

	if (!EAT) return NULL;

	DWORD ENTindex = -1;
	for (DWORD i = 0; i < exports.NumberOfNames; i++) {
		DWORD nameOffset = 0;
		if (!RVA2fileOffset (ENT[i], &nameOffset)) continue;
		char buf[128];
		if (!(readBytes (nameOffset, (PBYTE)buf, sizeof (buf)))) {
			_debug ("Cannot read at offset %x, err = %x", nameOffset, GetLastError());
			continue;
		}
		if (strlen(symName) < sizeof(buf)) buf[strlen(symName)] = '\0';
		else buf[sizeof(buf)-1] = '\0';
		if (!strncmp (buf, symName, sizeof (buf))) {
			ENTindex = i;
			break;
		}
	}

	if (ENTindex == -1) return NULL;
		
	DWORD ordinal = EOT[ENTindex];

	if (ordinal < 0 || ordinal >= exports.NumberOfFunctions) return NULL;

	return EAT[ordinal];
}
