#include "common.h"
#include "report.h"
#include "debug.h"
#include <windows.h>
#include <new>

#include <stdio.h>

SVVReport::SVVReport () {
	modInfo = NULL;
	discrBlk = NULL;
	sampleFilePaths = NULL;
	memset (&header, 0, sizeof (header));
	memset (&eor, 0, sizeof (eor));
	noOfSampleFiles = 0;
	sampleFilePaths = NULL;
}

SVVReport::~SVVReport() {

	delete [] modInfo;
	if (discrBlk) {
		for (DWORD i = 0; i < header.noOfModules; i++) delete []discrBlk[i];
		delete [] discrBlk;
	}
	
	for (DWORD i = 0; i < noOfSampleFiles; i++) delete sampleFilePaths[i];
	if (sampleFilePaths) delete [] sampleFilePaths;
}


bool SVVReport::prepare(SystemVirginityVerifier *svv) {

	outBufferSz = 0;

	strncpy (header.magic, RPTHDRMAGIC, sizeof (header.magic));
	header.version = RPTFMTVER;
	
	GetSystemTime (&header.timestamp);
	GetSystemInfo (&header.sysInfo);
	header.osVerInfo.dwOSVersionInfoSize = sizeof (header.osVerInfo);
	if (!GetVersionEx ((LPOSVERSIONINFO)&header.osVerInfo)) return false;

	header.noOfModules = svv->modules.count;


	header.currKiServiceTblStart = svv->kernelAgent->currKiServiceTblStart;
	header.currKiServiceTblEnd = svv->kernelAgent->currKiServiceTblEnd;
	header.origKiServiceTblStart = svv->kernelAgent->origKiServiceTblStart;
	header.origKiServiceTblEnd = svv->kernelAgent->origKiServiceTblEnd;

	outBufferSz += sizeof (HEADER);

	

	modInfo = new (std::nothrow) MODINFO [header.noOfModules];

	if (!modInfo) {
		SetLastError (ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}


	discrBlk = new (std::nothrow) PDISCRBLK [header.noOfModules];
	if (!discrBlk) {
		SetLastError (ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}

	for (DWORD i = 0; i < header.noOfModules; i++) {
		MODINFO *mi = &modInfo[i];
		Module *m = svv->modules.ptrTbl[i];

		mi->base = m->base;
		mi->errCode = m->errCode;
		mi->moduleImageHidden = m->moduleImageHidden;
		mi->moduleImageNotPresent = m->moduleImageNotPresent;
		mi->moduleImageReadError = m->moduleImageReadError;
		mi->size = m->size;
	

		mi->importantModule = m->importantModule;
		mi->kernelModule = m->kernelModule;
		mi->verdictLevel = m->verdictLevel;
		mi->isNtoskrnl = m->isNtoskrnl;

		wcscpy (mi->name, m->getName());
		wcscpy (mi->path, m->getPath());
		wcscpy (mi->effPath, m->getEffectivePath());

		outBufferSz += sizeof (MODINFO);

	
		mi->noOfDiscr = 0;
		if (!m->importantModule) continue;
		
		struct Module::DiscrepancyBlock *mDiscrBlk = m->discrBlocksHead;
		for (;mDiscrBlk; mDiscrBlk = mDiscrBlk->next) mi->noOfDiscr ++;
		if (mi->noOfDiscr) _debug ("module %S, noOfDiscr = %d", mi->name, mi->noOfDiscr);
		
		discrBlk[i] = new (std::nothrow) DISCRBLK [mi->noOfDiscr];
		if (!discrBlk[i]) {
			SetLastError (ERROR_NOT_ENOUGH_MEMORY);
			return false;
		}

		mDiscrBlk = m->discrBlocksHead;
		for (DWORD j = 0; mDiscrBlk; mDiscrBlk = mDiscrBlk->next, j++) {
			PDISCRBLK db = &discrBlk[i][j];


			memset (db, 0, sizeof (DISCRBLK));

			if (mDiscrBlk->section) 
				memcpy (&db->section, mDiscrBlk->section, sizeof (struct Module::Section));

			db->address = mDiscrBlk->address;
			db->len = mDiscrBlk->len;
			db->targetAddr = mDiscrBlk->targetAddr;
			db->verdictLevel = mDiscrBlk->verdictLevel;


			db->nearExportedSymbol = mDiscrBlk->nearExportedSymbol;
			db->offsetFromSymbol = mDiscrBlk->offsetFromSymbol;
			if (mDiscrBlk->exportedName) strncpy (db->exportedName, 
				mDiscrBlk->exportedName, sizeof (db->exportedName));


			db->type = mDiscrBlk->type;



			if (db->type == CODESECTION) {

				db->modification.code.isJMPingCode = mDiscrBlk->modification.code.isJMPingCode;
				db->modification.code.excludedByFilter = mDiscrBlk->modification.code.excludedByFilter;


				if (mDiscrBlk->modification.code.exlucsionDescription) wcsncpy (
					db->modification.code.exlucsionDescription, 
					mDiscrBlk->modification.code.exlucsionDescription, 
					sizeof (db->modification.code.exlucsionDescription)/sizeof (WCHAR));
				else db->modification.code.exlucsionDescription[0] = '\0';
			}


			if (db->type == SDT) {
				db->modification.SDT.ServTblEntrNo = mDiscrBlk->modification.SDT.ServTblEntrNo;


			}

			if (db->type == IDT) {
				db->modification.IDT.IdtEntryNo = mDiscrBlk->modification.IDT.IdtEntryNo;
				outBufferSz += sizeof (DISCRBLK);
				continue;			
			}

			if (db->type == MSR_SYSENTER) {
				outBufferSz += sizeof (DISCRBLK);
				continue;			
			}

			DWORD bytesQuoted = 
				(db->len > QUOTED_BYTES) ?
				QUOTED_BYTES : db->len;

			memcpy (db->inMemoryBytes,
				&mDiscrBlk->section->sectionInMemory[db->address],
				bytesQuoted);

			memcpy (db->inFileBytes,
				&mDiscrBlk->section->sectionInFile[db->address],
				bytesQuoted);

			outBufferSz += sizeof (DISCRBLK);

         }
	
	}

	noOfSampleFiles = 0;
	for (DWORD i = 0; i < header.noOfModules; i++) {
		Module *m = svv->modules.ptrTbl[i];
		if (includeModuleImage(m)) noOfSampleFiles ++;
		
	}

	_debug ("noOfSampleFiles = %d", noOfSampleFiles);

	sampleFilePaths = new (std::nothrow) PWCHAR [noOfSampleFiles];
	if (!sampleFilePaths) {
		SetLastError (ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}

	strcpy (eor.magic, RPTEORMAGIC);
	// eor.crc = 0;
	outBufferSz += sizeof (eor);



	DWORD noOfSample = 0;
	for (DWORD i = 0; i < header.noOfModules; i++) {
		Module *m = svv->modules.ptrTbl[i];
		
		_debug ("module %#x, %S (isImp = %d)", m, m->getName(), m->importantModule);

		if (!includeModuleImage(m)) continue;
		_debug (" including in the report");
		sampleFilePaths[noOfSample] = new (std::nothrow) WCHAR [wcslen (m->getEffectivePath()) + 1];
		
		if (!sampleFilePaths[noOfSample]) {
			SetLastError (ERROR_NOT_ENOUGH_MEMORY);
			return false;
		}
		
		wcscpy (sampleFilePaths[noOfSample], m->getEffectivePath());
		noOfSample ++;
	}

	return true;
}


bool SVVReport::includeModuleImage (Module *m) {
//	if (!m->importantModule) return false;
	return m->includeInReport;
//	if (!(m->moduleStatus == Module::MODULE_CLEAN && m->cleanLevel <= Module::CLEAN_LEVEL_EXCLUSIONS))
//		return true;
//	return false;
}


bool SVVReport::writeToBuf () {

	_debug ("allocating buffer for report (%d bytes)", outBufferSz);
	outBuffer = new (std::nothrow) BYTE [outBufferSz];
	if (!outBuffer) {
		_debug ("No memory for buffer!");
		SetLastError (ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}

	PBYTE ptr = outBuffer;
	_debug ("putting header at offset %d", ptr - outBuffer);
	memcpy (ptr, &header, sizeof (header));
	ptr += sizeof (header);

	for (DWORD i = 0; i < header.noOfModules; i++) {
		_debug ("putting modInfo[%d] at offset %d", i, ptr - outBuffer);
		memcpy (ptr, &modInfo[i], sizeof (MODINFO));
		ptr += sizeof (MODINFO);
	
		for (DWORD j = 0; j < modInfo[i].noOfDiscr; j++) {
			_debug ("putting discrBlk[%d][%d] at offset %d", i, j, ptr - outBuffer);
			memcpy (ptr, &discrBlk[i][j], sizeof (DISCRBLK));
			ptr += sizeof (DISCRBLK);
		}
	}


	_debug ("putting EOR at offset %d", ptr - outBuffer);
	memcpy (ptr, &eor, sizeof (eor));
	return true;
}


// WARNING: this assumes that all structers like modInfo are not allocated
bool SVVReport::extract (PBYTE buf, DWORD bufSz) {

	PBYTE ptr = buf;
	_debug ("reading header at offset %d", ptr - buf);
	memcpy (&header, ptr, sizeof (header));
	ptr += sizeof (header);


	if (0 != strcmp (header.magic, RPTHDRMAGIC)) {
		SetLastError (ERROR_BAD_FORMAT);
		return false;
	}

	if (header.version != RPTFMTVER) {
		SetLastError (ERROR_BAD_FORMAT);
		return false;
	}

	if (header.noOfModules > 1000) {
		_debug ("header.noOfModules > 1000!!!");
		SetLastError (ERROR_BAD_FORMAT);
		return false;
	}

    modInfo = new (std::nothrow) MODINFO [header.noOfModules];
	if (!modInfo) {
		SetLastError (ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}

	discrBlk = new (std::nothrow) PDISCRBLK [header.noOfModules];
	if (!discrBlk) {
		SetLastError (ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}

	for (DWORD i = 0; i < header.noOfModules; i++) {
		_debug ("reading modInfo[%d] at offset %d", i, ptr - buf);
		memcpy (&modInfo[i], ptr, sizeof (MODINFO));
		ptr += sizeof (MODINFO);

		_debug ("modInfo[i].noOfDiscr  = %d", modInfo[i].noOfDiscr);

		discrBlk[i] = new (std::nothrow) DISCRBLK [modInfo[i].noOfDiscr];
		if (!discrBlk[i]) {
			SetLastError (ERROR_NOT_ENOUGH_MEMORY);
			return false;
		}

		for (DWORD j = 0; j < modInfo[i].noOfDiscr; j++) {
			_debug ("reading discrBlk[%d][%d] at offset %d", i, j, ptr - buf);
			memcpy (&discrBlk[i][j], ptr, sizeof (DISCRBLK));
			ptr += sizeof (DISCRBLK);
		}


	}

	_debug ("reading EOR at offset %d", ptr - buf);
	memcpy (&eor, ptr, sizeof (eor));

	if (0 != strcmp (eor.magic, RPTEORMAGIC)) {
		_debug ("Wrong EOR magic!");
		SetLastError (ERROR_BAD_FORMAT);
		return false;
	}

	return true;
}

SVVReport::MODINFO* SVVReport::findModuleByAddr (DWORD address) {

	for (DWORD i = 0; i < header.noOfModules; i++) {
		if (address >= modInfo[i].base && 
			address < modInfo[i].base + modInfo[i].size) return &modInfo[i];

	}


	return NULL;
}

DWORD SVVReport::getReportVersion (PBYTE buf, DWORD bufSz) {

	if (0 != strcmp (header.magic, RPTHDRMAGIC)) return -1;

	if (bufSz < sizeof (header.magic) + sizeof (header.version)) return -1;
	HEADER *hdr = (HEADER*) buf;
	return hdr->version;

}

DWORD SVVReport::getSupportedVersion () {
	return RPTFMTVER;
}
