#include "svv.h"
#include "kernelagent.h"
#include "debug.h"
#include "ntdll.h"
#include <windows.h>
#include <psapi.h>

#include <new>
#include "idt.h"

SystemVirginityVerifier::SystemVirginityVerifier () {
	kernelModules.tbl = NULL;
	currProcModules.tbl = NULL;
	ntoskrnlModule = NULL;
	kernelAgent = NULL;
	diagnosticMode = false;
	
}

bool SystemVirginityVerifier::init (WCHAR *abrdDriverPath) {
	SYSTEM_INFO sysInfo;
	GetSystemInfo (&sysInfo);
	if (sysInfo.wProcessorArchitecture != PROCESSOR_ARCHITECTURE_INTEL) {
		_debug ("Not supported processor architecture!\n");
		SetLastError (ERROR_NOT_SUPPORTED);
		return false;
	}
		
	kernelAgent = new (std::nothrow) KernelAgent (abrdDriverPath);
	if (!kernelAgent) {
		SetLastError (ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}

	if (!kernelAgent->init()) return false;
	loadMandatoryDLLs();

	
	// we need a little bit more of power ;)
	// (especially for modules fixing)
	HANDLE pmy = GetCurrentProcess();
	TOKEN_PRIVILEGES NewState;

	NewState.PrivilegeCount=1;
	NewState.Privileges[0].Attributes=2;
	NewState.Privileges[0].Luid.HighPart=0;
	NewState.Privileges[0].Luid.LowPart=0;
	LookupPrivilegeValue(0,SE_DEBUG_NAME,&NewState.Privileges[0].Luid);
	HANDLE hToken;
	OpenProcessToken(pmy,0x20,&hToken);
	BOOL ret = AdjustTokenPrivileges(hToken,0,&NewState,0x10,0,0);
	CloseHandle(hToken);

	if (!ret) {
		_debug ("AdjustTokenPrivileges() failed!");
		return false;
	}


	WCHAR *impKernelMods[] = {
		L"ntoskrnl.exe",
		L"ftdisk.sys",
		L"ntfs.sys",
		L"disk.sys",
		L"tcpip.sys",
		L"ndis.sys",
		L"afd.sys",
		L"tdi.sys",


	};
	noOfImpKernelMods = sizeof(impKernelMods)/sizeof (WCHAR*);

	WCHAR *impUsermodeMods[] = {
		L"ntdll.dll",
		L"kernel32.dll",
		L"advapi32.dll",
		L"psapi.dll",
		L"ws2_32.dll",
		L"user32.dll"
		
	};
	noOfImpUsermodeMods = sizeof(impUsermodeMods)/sizeof (WCHAR*);

	impKrnlModPresent = new (std::nothrow) bool [noOfImpKernelMods];
	importantKernelModuleNames = new (std::nothrow) WCHAR* [noOfImpKernelMods];
	if (!importantKernelModuleNames) {
		SetLastError (ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}

	for (DWORD i = 0; i < noOfImpKernelMods; i++) {
		importantKernelModuleNames[i] = 
			new (std::nothrow) WCHAR [wcslen(impKernelMods[i]) + sizeof(WCHAR)];
		if (!importantKernelModuleNames[i]) {
			SetLastError (ERROR_NOT_ENOUGH_MEMORY);
			return false;
		}
        wcscpy (importantKernelModuleNames[i], impKernelMods[i]);
	}

	impUsrmdModPresent = new (std::nothrow) bool [noOfImpUsermodeMods];
	importantUsermodeModuleNames = new (std::nothrow) WCHAR* [noOfImpUsermodeMods];
	if (!importantUsermodeModuleNames) {
		SetLastError (ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}

	for (DWORD i = 0; i < noOfImpUsermodeMods; i++) {
		importantUsermodeModuleNames[i] = 
			new (std::nothrow) WCHAR [wcslen(impUsermodeMods[i]) + sizeof(WCHAR)];
		if (!importantUsermodeModuleNames[i]) {
			SetLastError (ERROR_NOT_ENOUGH_MEMORY);
			return false;
		}
		wcscpy (importantUsermodeModuleNames[i], impUsermodeMods[i]);
	}

	return true;
}

SystemVirginityVerifier::~SystemVirginityVerifier() {
	delete []kernelModules.tbl;
	delete []currProcModules.tbl;
	delete kernelAgent;


	for (DWORD i = 0; i < noOfImpKernelMods; i++)
		delete [](importantKernelModuleNames[i]);
	delete []importantKernelModuleNames;
	delete []impKrnlModPresent;

	for (DWORD i = 0; i < noOfImpUsermodeMods; i++)
		delete [](importantUsermodeModuleNames[i]);
	delete []importantUsermodeModuleNames;
	delete []impUsrmdModPresent;

}




bool SystemVirginityVerifier::isKernelModuleImportant(DWORD base) {

	if (base < kernelAgent->memoryStatusBlock.ullTotalVirtual)
			return false;	// skip NTDLL.DLL

	WCHAR baseName [MAX_FNAME_LEN];
	DWORD ret = GetDeviceDriverBaseName (
		(PVOID)base, baseName, sizeof(baseName));
	if (!ret) return false;

	for (DWORD j = 0; j < noOfImpKernelMods; j++)
		if (!_wcsicmp (importantKernelModuleNames[j], baseName)) {
			impKrnlModPresent [j] = true;
			return true;
			}
	return false;
}

bool SystemVirginityVerifier::getListOfKernelModules () {
	DWORD moduleBufSz = 0;

	DWORD status;
	if (!loadNtDllFunctions()) {
		return false;
	}

	status = ntQuerySystemInformation (
		SystemModuleInformation,
		&moduleBufSz, 0, &moduleBufSz);

	if (status != STATUS_INFO_LENGTH_MISMATCH) {
		_debug ("NtQuerySystemInformation () failed with status = %#x", status);
		SetLastError (ERROR_UNHANDLED_ERROR);	
		return false;
	}

	PBYTE moduleBuff = new (std::nothrow) BYTE [moduleBufSz];
	if (!moduleBuff) {
		SetLastError (ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}

	status = ntQuerySystemInformation (
		SystemModuleInformation,
		moduleBuff, moduleBufSz, 0);

	if (status != STATUS_SUCCESS) {
		_debug ("NtQuerySystemInformation (2nd call) failed with status = %#x", status);
		SetLastError (ERROR_UNHANDLED_ERROR);	
		return false;
	}

	DWORD moduleCount = *((DWORD*)(&moduleBuff[0]));
	_debug ("moduleCount = %d", moduleCount);

	for (DWORD i = 0; i < noOfImpKernelMods; i++) impKrnlModPresent[i] = false;

	PSYSTEM_MODULE_INFORMATION modInfo = 
		(PSYSTEM_MODULE_INFORMATION) (moduleBuff + sizeof (DWORD));

	kernelModules.tbl = new (std::nothrow) Module [moduleCount];

	for (DWORD i = 0; i < moduleCount; i++) {

		DWORD base = (DWORD)modInfo[i].base;
		kernelModules.tbl[i].setKernelAgentObject (kernelAgent);
		kernelModules.tbl[i].setBase (base);
		kernelModules.tbl[i].setSize (modInfo[i].size);
		WCHAR baseName [MAX_FNAME_LEN];

		DWORD ret = GetDeviceDriverBaseName (modInfo[i].base, baseName, sizeof(baseName));
		if (!ret) return false;
	
		kernelModules.tbl[i].setName (baseName);

		WCHAR fileName [MAX_PATH_LEN];
		ret = GetDeviceDriverFileName (modInfo[i].base, fileName, sizeof(fileName));
		if (!ret) return false;
	
		kernelModules.tbl[i].setPath (fileName);

		kernelModules.tbl[i].importantModule = isKernelModuleImportant (base);
		kernelModules.tbl[i].kernelModule = true;

		if (i == 0) {
			ntoskrnlModule = &kernelModules.tbl[i];
			kernelModules.tbl[i].isNtoskrnl = true;
		}

		kernelModules.tbl[i].diagnosticMode = diagnosticMode;
	}

	// now check if all important modules are present
	for (DWORD i = 0; i < noOfImpKernelMods; i++) 
		if (!impKrnlModPresent[i]) {
			SetLastError (ERROR_FILE_NOT_FOUND);
			return false;
		}
	_debug ("all important modules present");

	kernelModules.count = moduleCount;
	delete moduleBuff;


	if (!kernelAgent->getSDT()) {
		_debug ("Cannot get SDT!");
		return false;
	}


	DWORD testSectionNo = -1;
	for (DWORD i = 0; i < ntoskrnlModule->noOfSections; i++)
		if (!strcmp (ntoskrnlModule->sections[i].name, ".text")) testSectionNo = i;

	if (testSectionNo < 0) {
		_debug ("Can't get ntoskrnl's .text section!");
		return false;

	}




	DWORD kernelCodeStart = ntoskrnlModule->base + ntoskrnlModule->sections[testSectionNo].RVA;
	DWORD kernelCodeEnd = ntoskrnlModule->base + 
		ntoskrnlModule->sections[testSectionNo].RVA + ntoskrnlModule->sections[testSectionNo].size;
	
	if (kernelAgent->currKiServiceTblStart >= kernelCodeStart &&
		kernelAgent->currKiServiceTblStart < kernelCodeEnd) {

		kernelAgent->origKiServiceTblStart =
			kernelAgent->currKiServiceTblStart ;
		kernelAgent->origKiServiceTblEnd =
			kernelAgent->currKiServiceTblEnd;
        		
		
		} else {


			PEFile ntpefile (ntoskrnlModule->hImage);
			if (ntpefile.load() && ntpefile.loadRelocTable()) {
				kernelAgent->origKiServiceTblStart =
					findKiServiceTableRVA(&ntpefile) + ntoskrnlModule->base;
				kernelAgent->origKiServiceTblEnd =
					kernelAgent->origKiServiceTblStart + 0x500;
		        
			} else {
				return false;
			}

		}

    return true;
}

bool SystemVirginityVerifier::isUsermodeModuleImportant(DWORD base) {
	WCHAR modName [MAX_FNAME_LEN];
	DWORD ret = GetModuleBaseName (
		GetCurrentProcess(),
		(HMODULE)base,
		modName,
		sizeof (modName));
	
	if (!ret) return false;
	for (DWORD j = 0; j < noOfImpUsermodeMods; j++)
		if (!_wcsicmp (importantUsermodeModuleNames[j], modName)) {
				impUsrmdModPresent[j] = true;
				return true;
			}
	return false;
}

bool SystemVirginityVerifier::getListOfUsermodeModulesInCurrentProcess () {
	DWORD count = 100, neededSize;
	HMODULE *hMods = new (std::nothrow) HMODULE [count];
	BOOL ret;

	while (1) {
		ret = EnumProcessModules (
			GetCurrentProcess(),
			hMods,
			count * sizeof (HMODULE),
			&neededSize);

		if (!ret) return false;	

		if (neededSize > (count * sizeof (HMODULE))) {
			delete hMods;
			hMods = new (std::nothrow) HMODULE [count *= 2];
			continue;
		} else break;
	};


	DWORD noOfModules = neededSize/sizeof (HMODULE);

	for (DWORD i = 0; i < noOfImpUsermodeMods; i++) impUsrmdModPresent[i] = false;

	_debug ("noOfUsermodeModules = %d", noOfModules);
	currProcModules.tbl = new (std::nothrow) Module [noOfModules];

	for (DWORD i = 0; i < noOfModules; i++) {

		DWORD base = (DWORD)hMods[i];
		currProcModules.tbl[i].setKernelAgentObject (kernelAgent);
		currProcModules.tbl[i].setBase (base);

		WCHAR modName [MAX_FNAME_LEN];
		DWORD ret = GetModuleBaseName (
			GetCurrentProcess(),
			hMods[i],
			modName,
			sizeof (modName));
	
		if (!ret) return false;
		
		currProcModules.tbl[i].setName (modName);

		WCHAR modPath [MAX_PATH_LEN];
		ret = GetModuleFileName (
			hMods[i],
			modPath,
			sizeof (modPath));
	
		if (!ret) return false;
		currProcModules.tbl[i].setPath (modPath);


		MODULEINFO modInfo;
		ret = GetModuleInformation (GetCurrentProcess(),
			hMods[i], &modInfo,
			sizeof (modInfo));
		if (!ret) return false;

		currProcModules.tbl[i].size = modInfo.SizeOfImage;


		currProcModules.tbl[i].importantModule = isUsermodeModuleImportant (base);
	
	}

	// now check if all important modules are present
	for (DWORD i = 0; i < noOfImpUsermodeMods; i++) 
		if (!impUsrmdModPresent[i]) {
			SetLastError (ERROR_FILE_NOT_FOUND);
			return false;
	}


	delete []hMods;
	currProcModules.count = noOfModules;
    return true;
}

void SystemVirginityVerifier::createOneListOfModules () {
	modules.count = kernelModules.count + currProcModules.count;	
	modules.ptrTbl = new (std::nothrow) Module* [modules.count];

	for (DWORD i = 0; i < kernelModules.count; i++)
		modules.ptrTbl[i] = &kernelModules.tbl[i];

	for (DWORD i = 0; i < currProcModules.count; i++)
		modules.ptrTbl[kernelModules.count + i] =
		&currProcModules.tbl[i];

}

Module* SystemVirginityVerifier::findModuleByAddr (DWORD address) {
	if (modules.count == 0 || modules.ptrTbl == NULL) {
		_debug ("modules is NULL!");
		return NULL;
	}

	
	for (DWORD i = 0; i < modules.count; i++) {
		Module *m = modules.ptrTbl[i];
		if (address >= m->base && address < m->base + m->size)
			return m;
	}
	return NULL;
}
	

bool SystemVirginityVerifier::loadDriver () {
	
	if (!kernelAgent) return false;

	if (kernelAgent->isDriverOk()) return true;
	
	return kernelAgent->loadDriver();
}

bool SystemVirginityVerifier::unloadDriver () {

	if (!kernelAgent) return false;

	return kernelAgent->unloadDriver();
}

bool SystemVirginityVerifier::loadMandatoryDLLs() {

	// using LoadLibrary() is not a good idea here...

	RegOpenKey (0, 0, 0);	// ADVAPI32.DLL
	WSAStartup (0, 0);		// WSOCK32.DLL
	GetWindow (0, 0);		// USER32.DLL

	//...
	return true;
}


SystemVerdict SystemVirginityVerifier::verdict (bool *disinfectionPossible) {
	
	SystemVerdict totalWorstVerdict = INVALID_VERDICT;
	*disinfectionPossible = true;

	if (kernelAgent->currKiServiceTblStart != kernelAgent->origKiServiceTblStart) 
		// this might be Kaspersky
		totalWorstVerdict = VERDICT_YELLOW;


	for (DWORD i = 0; i < modules.count; i++) {
		
		Module *m = modules.ptrTbl[i];
		m->verdictLevel = VERDICT_BLUE;	// let's be optimistic at first :)
		if (!m->importantModule) continue;

		if (m->errCode) {
			if (m->importantModule) *disinfectionPossible = false;
			m->verdictLevel = VERDICT_DEEPRED;
			goto skip_testing_this_module;
		}

		if (m->moduleImageNotPresent || m->moduleImageReadError || m->moduleImageHidden) {
			if (m->importantModule) *disinfectionPossible = false;
			m->verdictLevel = VERDICT_DEEPRED;
			goto skip_testing_this_module;
		}
		
		
		struct Module::DiscrepancyBlock *discrBlk = m->discrBlocksHead;

		for (;discrBlk; discrBlk = discrBlk->next) {
		

			if (discrBlk->verdictLevel > 0) {
				if (m->verdictLevel < discrBlk->verdictLevel)
					m->verdictLevel = discrBlk->verdictLevel;
				continue;
			}


			if (discrBlk->type == CODESECTION && discrBlk->modification.code.excludedByFilter) {
				if (m->verdictLevel < VERDICT_GREEN)
					m->verdictLevel = VERDICT_GREEN;
				discrBlk->verdictLevel = VERDICT_GREEN;
				continue;
			}
			if (discrBlk->type == EAT) {
				if (m->verdictLevel < VERDICT_YELLOW)
					m->verdictLevel = VERDICT_YELLOW;
				discrBlk->verdictLevel = VERDICT_YELLOW;
				continue;				// TODO: heurisitics here
			}

			if (discrBlk->targetAddr) {
									
				Module *targetModule = findModuleByAddr (discrBlk->targetAddr);
				if (!targetModule) {
					// good programs do not have code in anonymous areas
					if (m->verdictLevel < VERDICT_DEEPRED)
						m->verdictLevel = VERDICT_DEEPRED;
					m->includeInReport = true;
					discrBlk->verdictLevel = VERDICT_DEEPRED;
					continue;
				}
				
				
				if (targetModule->moduleImageHidden) {
					// rootkit for 100%
					if (m->verdictLevel < VERDICT_DEEPRED)
						m->verdictLevel = VERDICT_DEEPRED;						
					m->includeInReport = true;
					targetModule->includeInReport = true;
					discrBlk->verdictLevel = VERDICT_DEEPRED;
					continue;
				}
				
				if (targetModule->moduleImageNotPresent) {
					// This may be DbgView or RegMon but also some rootkit, like Vanquish

					WCHAR lwcstr[MAX_PATH_LEN];
					wcscpy (lwcstr, targetModule->getPath());
					_wcslwr (lwcstr);
					if (wcsstr (lwcstr, L"system32\\drivers") == NULL) {
					// this doesn't look like an innocent tool 
						if (m->verdictLevel < VERDICT_DEEPRED)
							m->verdictLevel = VERDICT_DEEPRED;	
						discrBlk->verdictLevel = VERDICT_DEEPRED;
						m->includeInReport = true;
						continue;
					} else {
					// probbaly some kind of tracing tool
						if (m->verdictLevel < VERDICT_ORANGE)
							m->verdictLevel = VERDICT_ORANGE;
						m->includeInReport = true;
						discrBlk->verdictLevel = VERDICT_ORANGE;
						continue;	
					}
				}

				// module is visible and present, probbaly personal firewall or so
				if (m->verdictLevel < VERDICT_YELLOW)
					m->verdictLevel = VERDICT_YELLOW;
				discrBlk->verdictLevel = VERDICT_YELLOW;


			} else { // discrBlk->targetAddr = NULL
				//if (!wcsncmp (m->getName(), L"ntoskrnl", 8)) {
				if (m == ntoskrnlModule) {
					// some unkown modifications in kernel
					// this may be false positive
					// be careful with fix this!!!
					if (m->verdictLevel < VERDICT_RED)
						m->verdictLevel = VERDICT_RED;				
					m->includeInReport = true;
					discrBlk->verdictLevel = VERDICT_RED;
					continue;
				} else { // not in ntoskrnl
					// unkown modification (not JMP nor SDT hook)
					if (m->verdictLevel < VERDICT_DEEPRED)
						m->verdictLevel = VERDICT_DEEPRED;						
					m->includeInReport = true;
					discrBlk->verdictLevel = VERDICT_DEEPRED;
					continue;				
				}
			}

		}	// for (;discrBlk; 

	skip_testing_this_module:
		if (m->importantModule) {
			if (totalWorstVerdict < m->verdictLevel)
			totalWorstVerdict = m->verdictLevel;
		}

	}

	return totalWorstVerdict;
}

// credits to 90210 :)
DWORD SystemVirginityVerifier::findKiServiceTableRVA (PEFile *ntpe) {

	if (!ntpe) return NULL;
		
	DWORD SDTRVA = ntpe->findExportedSymbol ("KeServiceDescriptorTable");

	_debug ("KeServiceDescriptorTable RVA = %#x", SDTRVA);
	PBYTE relocTable = ntpe->relocTable;
	DWORD relocTableSz = ntpe->relocTableSz;

	DWORD boundBase = ntpe->ntHdr.OptionalHeader.ImageBase;

	_debug ("boundBase = %#x", boundBase);
	int relEntNo = 0;

	for (IMAGE_BASE_RELOCATION *relEntry = (IMAGE_BASE_RELOCATION *)relocTable;
		(((DWORD)relEntry) < ((DWORD)relocTable + relocTableSz) &&
		relEntry->SizeOfBlock);
		relEntry = (IMAGE_BASE_RELOCATION *)(((DWORD)relEntry) + relEntry->SizeOfBlock)) {
	

			WORD *relWordTbl = (WORD*)(((DWORD)relEntry) + 0x8);
			int relWordTblSz = (relEntry->SizeOfBlock - 0x8)/sizeof (WORD);
			
//			_debug ("relEntNo: %d, relWordTblSz: %d, baseRVA = %#x", relEntNo++, relWordTblSz, relEntry->VirtualAddress);
			
			for (int i = 0; i < relWordTblSz; i++)  {
		
				DWORD addr = relEntry->VirtualAddress + 
					(relWordTbl[i] & 0x0fff);

			

				DWORD relContents;
				DWORD KiServTblRVA = 0;
				BYTE opcode[2];

				switch ((relWordTbl[i] & 0xf000)>>12) {
					case IMAGE_REL_BASED_ABSOLUTE:
						break;
					case IMAGE_REL_BASED_HIGHLOW:
						// about to fixup SDT ?
						DWORD fileOffset;
						if (!ntpe->RVA2fileOffset (addr, &fileOffset)) break;
						
						ntpe->readBytes (fileOffset, (PBYTE)&relContents, sizeof (DWORD));
						
						//boundBase = ntpe->ntHdr.OptionalHeader.ImageBase;
						
						relContents -= boundBase;
				
						
						if (relContents == SDTRVA) {
							_debug (" SDTRVA found at RVA %#x (fileOffset = %#x)", addr, fileOffset);
							ntpe->readBytes (fileOffset - 2, (PBYTE)&opcode, sizeof (opcode));
							if (!memcmp (opcode, "\xc7\x05", 2)) {
								ntpe->readBytes (fileOffset + 4, (PBYTE)&KiServTblRVA, sizeof (DWORD));
								KiServTblRVA -= boundBase;
								_debug ("mov [%x], %x", 
									relContents, 
									KiServTblRVA);
								return KiServTblRVA;
							}

						}
						
						break;
					default: break;      
				}   			
			} 	 
		
		}



	return 0;


}





bool SystemVirginityVerifier::verifyKernelIDT () {

	_debug ("verifyKernelIDT()");

	kernelAgent->getIDT();
	_debug ("IDT base = %#x, sz = %#x", kernelAgent->IDTbaseVA, kernelAgent->IDTsize);

	

	BYTE *idtBuffer = new (std::nothrow) BYTE [kernelAgent->IDTsize];
	if (!idtBuffer) {
		SetLastError (ERROR_NOT_ENOUGH_MEMORY);
		return false;
	}

	DWORD bRead = 0;
	bool ret = kernelAgent->readVirtualMemory (
		kernelAgent->IDTbaseVA,
		idtBuffer,
		kernelAgent->IDTsize,
		&bRead);

	if (!ret || bRead != kernelAgent->IDTsize) {
		SetLastError (ERROR_INVALID_ADDRESS);	
		return false;
	}

	DWORD noOfEntries = kernelAgent->IDTsize/sizeof (IDTGATE);
	_debug ("IDT no of entries: %#d", noOfEntries);

	PIDTGATE idtTbl = (PIDTGATE) idtBuffer;
	

	DWORD idtEntriesToCheck[] = {0, 1, 3, 4, 5, 6, 7, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 0x2e, 0xff };

	for (DWORD i = 0; i < sizeof (idtEntriesToCheck)/sizeof(DWORD); i++) {

		int entNo = idtEntriesToCheck[i];

		if ((idtTbl[entNo].flags & 0x7) == 0) {

			_debug (" IDT[%d] NULL", entNo);
			continue;
		}

		else if ((idtTbl[entNo].flags & 0x7) == 5) {
			
			_debug (" IDT[%d] TASK gate", entNo);
			continue;
		
		}
		else if ((idtTbl[entNo].flags & 0x7) == 6) {

			DWORD targetAddr = idtTbl[entNo].off1 + (idtTbl[entNo].off2<<16);
			Module *targetModule = findModuleByAddr (targetAddr);

			_debug (" IDT[%d] INT gate (%s) %#x (module %S)", 
			entNo, (idtTbl[entNo].flags & 0x8) ? "32bit" : "16bit",
			targetAddr, (targetModule) ? targetModule->getName() : L"UNKNOWN!");

		
			bool suspicious = true;
			if (targetModule && targetModule->isNtoskrnl) {
				
				for (DWORD s = 0; s < ntoskrnlModule->noOfSections; s++) 
					if (targetModule->sections[s].RVA + ntoskrnlModule->base <= targetAddr &&
						targetAddr < targetModule->sections[s].RVA + ntoskrnlModule->base + targetModule->sections[s].size) {
							_debug (" points to section %s", targetModule->sections[s].name); 
							suspicious = false;

						}
			}

			if (suspicious) {

				_debug ("suspicious!");
				struct Module::DiscrepancyBlock* discrBlk = new (std::nothrow) struct Module::DiscrepancyBlock;
				if (discrBlk == NULL) {
					SetLastError (ERROR_NOT_ENOUGH_MEMORY);
					return false;
				}
				memset (discrBlk, 0, sizeof (struct Module::DiscrepancyBlock));
				discrBlk->type = IDT;
							
				
				discrBlk->section = NULL;
		

				discrBlk->address = 0;
				discrBlk->len = 4;

				discrBlk->modification.IDT.IdtEntryNo = entNo;

				discrBlk->targetAddr = targetAddr;
				if (!targetModule)
					discrBlk->verdictLevel = VERDICT_DEEPRED;
				if (targetModule && !targetModule->isNtoskrnl)
					discrBlk->verdictLevel = VERDICT_DEEPRED;

				
				if (targetModule && targetModule->isNtoskrnl) {
					discrBlk->verdictLevel = VERDICT_RED;
			//		discrBlk->modification.IDT.targetNotInCodeSection = true;
				}
	
				// insert at the end (sorted by address)
				struct Module::DiscrepancyBlock* p = ntoskrnlModule->discrBlocksHead;
				if (!p) ntoskrnlModule->discrBlocksHead = discrBlk;
				else {
					for (; p->next; p = p->next);			
					p->next = discrBlk;
				}
		
			}



		}
		else if ((idtTbl[entNo].flags & 0x7) == 7) {

            _debug (" IDT[%d] TRAP gate (%s) %#x", 
			entNo, (idtTbl[entNo].flags & 0x8) ? "32bit" : "16bit",
			idtTbl[entNo].off1 + (idtTbl[entNo].off2<<16));
			continue;

		}
		else {
			_debug (" IDT[%d] - unkown flags = %#x", entNo, idtTbl[entNo].flags);

		}


	}






	return true;
}


bool SystemVirginityVerifier::verifyKernelMSRs () {

	_debug ("verifyKernelMSRs()");

	kernelAgent->getMSRs();

	_debug ("SYSENTER_EIP = %#x\n", kernelAgent->SYSENTER_EIP);

	Module *targetModule = findModuleByAddr (kernelAgent->SYSENTER_EIP);

	DWORD targetAddr = kernelAgent->SYSENTER_EIP;
	bool suspicious = true;
	if (targetModule && targetModule->isNtoskrnl) {
		
		for (DWORD s = 0; s < ntoskrnlModule->noOfSections; s++) 
			if (targetModule->sections[s].RVA + ntoskrnlModule->base <= targetAddr &&
				targetAddr < targetModule->sections[s].RVA + ntoskrnlModule->base + targetModule->sections[s].size) {
					_debug (" points to section %s", targetModule->sections[s].name); 
					suspicious = false;

				}
	}

	if (targetAddr == 0) suspicious = false;	// Windows 2000 systems do not use SYSENTER instruction
	if (suspicious) {

			struct Module::DiscrepancyBlock* discrBlk = new (std::nothrow) struct Module::DiscrepancyBlock;
			if (discrBlk == NULL) {
				SetLastError (ERROR_NOT_ENOUGH_MEMORY);
				return false;
			}
			memset (discrBlk, 0, sizeof (struct Module::DiscrepancyBlock));
			discrBlk->type = MSR_SYSENTER;
							
			discrBlk->targetAddr = targetAddr;

			if (!targetModule)
				discrBlk->verdictLevel = VERDICT_DEEPRED;
			if (targetModule && !targetModule->isNtoskrnl)
				discrBlk->verdictLevel = VERDICT_DEEPRED;
			if (targetModule && targetModule->isNtoskrnl)
				discrBlk->verdictLevel = VERDICT_RED;

			// insert at the end (sorted by address)
			struct Module::DiscrepancyBlock* p = ntoskrnlModule->discrBlocksHead;
			if (!p) ntoskrnlModule->discrBlocksHead = discrBlk;
			else {
				for (; p->next; p = p->next);			
				p->next = discrBlk;
			}
	}

	return true;
}
